from typing import Any, NoReturn, Optional
from typing import Union as _Union

from pipgrip.libs.mixology.union import Union


class Range(object):
    """
    A range of versions represented by a minimum version,
    a maximum version and whether they are included in the range.

    Both the minimum and maximum versions can be left out
    to represent unbounded ranges.

    A single version is represented by Range(version, version, True, True).
    """

    def __init__(
        self, min=None, max=None, include_min=False, include_max=False, string=None
    ):  # type: (Any, Any, bool, bool, Optional[str]) -> None
        self._min = min
        self._max = max
        self._include_min = include_min
        self._include_max = include_max
        self._hash = None
        self._string = string

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max

    @property
    def include_min(self):
        return self._include_min

    @property
    def include_max(self):
        return self._include_max

    @property
    def inverse(self):  # type: () -> _Union[Range, Union]
        if self.is_any():
            return EmptyRange()

        low = Range(max=self.min, include_max=not self.include_min)
        high = Range(min=self.max, include_min=not self.include_max)

        if self.min is None:
            return high

        if self.max is None:
            return low

        return low.union(high)

    def is_empty(self):
        return False

    def is_any(self):
        return self._min is None and self._max is None

    def allows_all(self, other):  # type: (Range) -> bool
        if other.is_empty():
            return True

        if isinstance(other, Union):
            return all([self.allows_all(constraint) for constraint in other.ranges])

        return not other.allows_lower(self) and not other.allows_higher(self)

    def allows_any(self, other):  # type: (Range) -> bool
        if other.is_empty():
            return False

        if isinstance(other, Union):
            return any([self.allows_any(constraint) for constraint in other.ranges])

        return not other.is_strictly_lower(self) and not other.is_strictly_higher(self)

    def intersect(self, other):  # type: (Range) -> Range
        if other.is_empty():
            return other

        if isinstance(other, Union):
            return other.intersect(self)

        if other.is_single_version():
            if self.allows_all(other):
                return other

            return EmptyRange()

        if self.allows_lower(other):
            if self.is_strictly_lower(other):
                return EmptyRange()

            intersect_min = other.min
            intersect_include_min = other.include_min
        else:
            if other.is_strictly_lower(self):
                return EmptyRange()

            intersect_min = self._min
            intersect_include_min = self._include_min

        if self.allows_higher(other):
            intersect_max = other.max
            intersect_include_max = other.include_max
        else:
            intersect_max = self._max
            intersect_include_max = self._include_max

        if intersect_min is None and intersect_max is None:
            return Range()

        # If the range is just a single version.
        if intersect_min == intersect_max:
            # Because we already verified that the lower range isn't strictly
            # lower, there must be some overlap.
            assert intersect_include_min and intersect_include_max

            return Range(intersect_min, intersect_max, True, True)

        # If we got here, there is an actual range.
        return Range(
            intersect_min, intersect_max, intersect_include_min, intersect_include_max
        )

    def union(self, other):  # type: (Range) -> _Union[Range, Union]
        if isinstance(other, Union):
            return other.union(self)

        if not self.is_contiguous_to(other):
            return Union.of(self, other)

        if self.allows_lower(other):
            union_min = self.min
            union_include_min = self.include_min
        else:
            union_min = other.min
            union_include_min = other.include_min

        if self.allows_higher(other):
            union_max = self.max
            union_include_max = self.include_max
        else:
            union_max = other.max
            union_include_max = other.include_max

        return Range(
            union_min,
            union_max,
            include_min=union_include_min,
            include_max=union_include_max,
        )

    def is_contiguous_to(self, other):  # type: (Range) -> bool
        if other.is_empty():
            return False

        return (
            self.allows_any(other)
            or (self.max == other.min and (self.include_max or other.include_min))
            or (self.min == other.max and (self.include_min or other.include_max))
        )

    def difference(self, other):  # type: (_Union[Range, Union]) -> _Union[Range, Union]
        if other.is_empty():
            return self

        if isinstance(other, Union):
            ranges = []
            current = self

            for range in other.ranges:
                # Skip any ranges that are strictly lower than [current].
                if range.is_strictly_lower(current):
                    continue

                # If we reach a range strictly higher than [current], no more ranges
                # will be relevant so we can bail early.
                if range.is_strictly_higher(current):
                    break

                difference = current.difference(range)
                if difference.is_empty():
                    return EmptyRange()
                elif isinstance(difference, Union):
                    # If [range] split [current] in half, we only need to continue
                    # checking future ranges against the latter half.
                    ranges.append(difference.ranges[0])
                    current = difference.ranges[-1]
                else:
                    current = difference

            if not ranges:
                return current

            return Union.of(*(ranges + [current]))

        if not self.allows_any(other):
            return self

        if not self.allows_lower(other):
            before = None
        elif self.min == other.min:
            before = Range(self.min, self.min, True, True)
        else:
            before = Range(self.min, other.min, self.include_min, not other.include_min)

        if not self.allows_higher(other):
            after = None
        elif self.max == other.max:
            after = Range(self.max, self.max, True, True)
        else:
            after = Range(other.max, self.max, not other.include_max, self.include_max)

        if before is None and after is None:
            return EmptyRange()

        if before is None:
            return after

        if after is None:
            return before

        return Union.of(before, after)

    def allows_lower(self, other):  # type: (Range) -> bool
        if self.min is None:
            return other.min is not None

        if other.min is None:
            return False

        if self.min < other.min:
            return True

        if self.min > other.min:
            return False

        return self.include_min and not other.include_min

    def allows_higher(self, other):  # type: (Range) -> bool
        if self.max is None:
            return other.max is not None

        if other.max is None:
            return False

        if self.max < other.max:
            return False

        if self.max > other.max:
            return True

        return self.include_max and not other.include_max

    def is_strictly_lower(self, other):  # type: (Range) -> bool
        if self.max is None or other.min is None:
            return False

        if self.max < other.min:
            return True

        if self.max > other.min:
            return False

        return not self.include_max or not other.include_min

    def is_strictly_higher(self, other):  # type: (Range) -> bool
        return other.is_strictly_lower(self)

    def is_adjacent_to(self, other):  # type: (Range) -> bool
        if self.max != other.min:
            return False

        return (
            self.include_max
            and not other.include_min
            or not self.include_max
            and other.include_min
        )

    def is_single_version(self):  # type: () -> bool
        return (
            self.min is not None
            and self.min == self.max
            and self.include_min
            and self.include_max
        )

    def __eq__(self, other):
        if not isinstance(other, Range):
            return False

        return (
            self._min == other.min
            and self._max == other.max
            and self._include_min == other.include_min
            and self._include_max == other.include_max
        )

    def __lt__(self, other):
        return self._cmp(other) < 0

    def __le__(self, other):
        return self._cmp(other) <= 0

    def __gt__(self, other):
        return self._cmp(other) > 0

    def __ge__(self, other):
        return self._cmp(other) >= 0

    def _cmp(self, other):  # type: (Range) -> int
        if self.min is None:
            if other.min is None:
                return self._compare_max(other)

            return -1
        elif other.min is None:
            return 1

        if self.min != other.min:
            return -1 if self.min < other.min else 1

        if self.include_min != other.include_min:
            return -1 if self.include_min else 1

        return self._compare_max(other)

    def _compare_max(self, other):  # type: (Range) -> int
        if self.max is None:
            if other.max is None:
                return 0

            return 1
        elif other.max is None:
            return -1

        if self.max != other.max:
            return 1 if self.max < other.max else -1

        if self.include_max != other.include_max:
            return 1 if self.include_max else -1

        return 0

    def __str__(self):
        if self._string is not None:
            return self._string

        text = ""

        if self.min is not None:
            if self.min == self.max and self.include_min and self.include_max:
                return "{}".format(self.min)

            text += ">=" if self.include_min else ">"
            text += self.min.text

        if self.max is not None:
            if self.min is not None:
                text += ","

            text += "{}{}".format("<=" if self.include_max else "<", self.max.text)

        if self.min is None and self.max is None:
            return "*"

        return text

    def __repr__(self):
        return "<Range ({})>".format(str(self))

    def __hash__(self):
        if self._hash is None:
            self._hash = (
                hash(self.min)
                ^ hash(self.max)
                ^ hash(self.include_min)
                ^ hash(self.include_max)
            )

        return self._hash


class EmptyRange(Range):
    @property
    def min(self):  # type: () -> NoReturn
        raise NotImplementedError()

    @property
    def max(self):  # type: () -> NoReturn
        raise NotImplementedError()

    @property
    def include_min(self):  # type: () -> NoReturn
        raise NotImplementedError()

    @property
    def include_max(self):  # type: () -> NoReturn
        raise NotImplementedError()

    def is_empty(self):  # type: () -> bool
        return True

    def is_any(self):  # type: () -> bool
        return False

    def __eq__(self, other):  # type: (Range) -> bool
        return other.is_empty()

    def intersect(self, other):  # type: (Range) -> Range
        return self

    def allows_all(self, other):  # type: (Range) -> bool
        return other.is_empty()

    def allows_any(self, other):  # type: (Range) -> bool
        return other.is_empty()

    def is_single_version(self):  # type: () -> bool
        return False

    @property
    def inverse(self):  # type: () -> _Union[Range, Union]
        return Range()

    def __str__(self):  # type: () -> str
        return "(no versions)"



from distutils.version import LooseVersion

import pip

try:
    from urllib.parse import urlparse
except ImportError:
    # Python 2
    from urlparse import urlparse  # noqa:F401

PIP_VERSION = LooseVersion(pip.__version__).version

if PIP_VERSION < [10]:
    from pip.locations import USER_CACHE_DIR
else:
    from pip._internal.locations import USER_CACHE_DIR  # noqa:F401



# -*- coding: utf-8 -*-
from typing import Hashable, Optional

from pipgrip.libs.mixology.constraint import Constraint
from pipgrip.libs.mixology.package import Package
from pipgrip.libs.mixology.range import EmptyRange
from pipgrip.libs.mixology.set_relation import SetRelation
from pipgrip.pipper import parse_req


class Term(object):
    """
    A statement about a package which is true or false for a given selection of
    package versions.
    """

    def __init__(self, constraint, is_positive):  # type: (Constraint, bool) -> None
        self._constraint = constraint
        self._package = constraint.package
        self._positive = is_positive
        self._normalized_constraint = None
        self._empty = None

    @property
    def inverse(self):  # type: () -> Term
        return Term(self.constraint, not self.is_positive())

    @property
    def package(self):  # type: () -> Hashable
        return self._package

    @property
    def constraint(self):  # type: () -> Constraint
        return self._constraint

    @property
    def normalized_constraint(self):  # type: () -> Constraint
        if self._normalized_constraint is None:
            self._normalized_constraint = (
                self.constraint if self.is_positive() else self.constraint.inverse
            )

        return self._normalized_constraint

    def is_positive(self):  # type: () -> bool
        return self._positive

    def satisfies(self, other):  # type: (Term) -> bool
        """
        Returns whether this term satisfies another.
        """
        return (
            self._package == other.package
            and self.relation(other) == SetRelation.SUBSET
        )

    def relation(self, other):  # type: (Term) -> SetRelation
        """
        Returns the relationship between the package versions
        allowed by this term and another.
        """
        if self.package != other.package:
            raise ValueError("{} should refer to {}".format(other, self.package))

        if other.is_positive():
            if self.is_positive():
                if not self.is_compatible_with(other):
                    return SetRelation.DISJOINT

                # foo ^1.5.0 is a subset of foo ^1.0.0
                if other.constraint.allows_all(self.constraint):
                    return SetRelation.SUBSET

                # foo ^2.0.0 is disjoint with foo ^1.0.0
                if not self.constraint.allows_any(other.constraint):
                    return SetRelation.DISJOINT

                return SetRelation.OVERLAPPING
            else:
                if not self.is_compatible_with(other):
                    return SetRelation.OVERLAPPING

                # not foo ^1.0.0 is disjoint with foo ^1.5.0
                if self.constraint.allows_all(other.constraint):
                    return SetRelation.DISJOINT

                # not foo ^1.5.0 overlaps foo ^1.0.0
                # not foo ^2.0.0 is a superset of foo ^1.5.0
                return SetRelation.OVERLAPPING
        else:
            if self.is_positive():
                if not self.is_compatible_with(other):
                    return SetRelation.SUBSET

                # foo ^2.0.0 is a subset of not foo ^1.0.0
                if not other.constraint.allows_any(self.constraint):
                    return SetRelation.SUBSET

                # foo ^1.5.0 is disjoint with not foo ^1.0.0
                if other.constraint.allows_all(self.constraint):
                    return SetRelation.DISJOINT

                # foo ^1.0.0 overlaps not foo ^1.5.0
                return SetRelation.OVERLAPPING
            else:
                if not self.is_compatible_with(other):
                    return SetRelation.OVERLAPPING

                # not foo ^1.0.0 is a subset of not foo ^1.5.0
                if self.constraint.allows_all(other.constraint):
                    return SetRelation.SUBSET

                # not foo ^2.0.0 overlaps not foo ^1.0.0
                # not foo ^1.5.0 is a superset of not foo ^1.0.0
                return SetRelation.OVERLAPPING

    def intersect(self, other):  # type: (Term) -> Term
        """
        Returns a Term that represents the packages
        allowed by both this term and another
        """
        if self.package != other.package:
            raise ValueError("{} should refer to {}".format(other, self.package))

        if self.is_compatible_with(other):
            if self.is_positive() != other.is_positive():
                # foo ^1.0.0 ∩ not foo ^1.5.0 → foo >=1.0.0 <1.5.0
                positive = self if self.is_positive() else other
                negative = other if self.is_positive() else self

                to_return = self._non_empty_term(
                    positive.constraint.difference(negative.constraint), True
                )
            elif self.is_positive():
                # foo ^1.0.0 ∩ foo >=1.5.0 <3.0.0 → foo ^1.5.0
                to_return = self._non_empty_term(
                    self.constraint.intersect(other.constraint), True
                )
            else:
                # not foo ^1.0.0 ∩ not foo >=1.5.0 <3.0.0 → not foo >=1.0.0 <3.0.0
                to_return = self._non_empty_term(
                    self.constraint.union(other.constraint), False
                )
            if to_return is not None:
                to_return._constraint._package._req = parse_req(
                    to_return.constraint.package.req.__str__(),
                    extras=self.constraint.package.req.extras
                    | other.constraint.package.req.extras,
                )
                to_return._package = self.constraint.package

        elif self.is_positive() != other.is_positive():
            to_return = self if self.is_positive() else other
        else:
            to_return = Term(Constraint(self.package, EmptyRange()))

        return to_return

    def difference(self, other):  # type: (Term) -> Term
        """
        Returns a Term that represents packages
        allowed by this term and not by the other
        """
        return self.intersect(other.inverse)

    def is_compatible_with(self, other):  # type: (Term) -> bool
        return (
            self.package == Package.root()
            or other.package == Package.root()
            or self.package == other.package
        )

    def is_empty(self):  # type: () -> bool
        if self._empty is None:
            self._empty = self.normalized_constraint.is_empty()

        return self._empty

    def _non_empty_term(
        self, constraint, is_positive
    ):  # type: (Constraint, bool) -> Optional[Term]
        if constraint.is_empty():
            return

        return Term(constraint, is_positive)

    def to_string(self, allow_every=False):  # type: (bool) -> str
        if self.is_positive():
            return self.constraint.to_string(allow_every=allow_every)

        return "not {}".format(self.constraint)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return "<Term {}>".format(str(self))



import pkg_resources

from pipgrip.pipper import parse_req


class Package(object):
    """Represent a project's package."""

    def __init__(self, pip_string):  # type: (str) -> None
        if pip_string == "_root_":
            req = parse_req(".")
            req.key = "_root_"
        else:
            req = parse_req(pip_string)
        self._name = req.key
        self._req = req

    @classmethod
    def root(cls):  # type: () -> Package
        return Package("_root_")

    @property
    def name(self):  # type: () -> str
        return self._name

    @property
    def req(self):  # type: () -> pkg_resources.Requirement
        return self._req

    def __eq__(self, other):  # type: () -> bool
        return str(other) == self.name

    def __ne__(self, other):  # type: () -> bool
        return not self.__eq__(other)

    def __str__(self):  # type: () -> str
        return self._name

    def __repr__(self):  # type: () -> str
        return 'Package("{}")'.format(self.req.extras_name)

    def __hash__(self):
        return hash(self.name)



import io
import logging
import os
import tempfile
from collections import OrderedDict
from json import dumps
from subprocess import CalledProcessError

import click
from anytree import Node, RenderTree
from anytree.search import findall_by_attr
from packaging.markers import default_environment

from pipgrip.compat import USER_CACHE_DIR
from pipgrip.libs.mixology.failure import SolverFailure
from pipgrip.libs.mixology.package import Package
from pipgrip.libs.mixology.version_solver import VersionSolver
from pipgrip.package_source import PackageSource
from pipgrip.pipper import install_packages, read_requirements

logging.basicConfig(format="%(levelname)s: %(message)s")
logger = logging.getLogger()


def flatten(d):
    out = OrderedDict()
    for key0, val0 in d.items():
        out[key0[0]] = key0[1]
        if not val0:
            continue
        for key1, subdict in val0.items():
            out[key1[0]] = key1[1]
            deeper = flatten(subdict).items()
            for key2, val2 in deeper:
                if key2 in out and out[key2] != val2:
                    raise RuntimeError(
                        "{} has not been solved: both {} and {} found... Please file an issue on GitHub.",
                        key2,
                        val0,
                        val2,
                    )
                else:
                    out[key2] = val2
    return out


def _find_version(source, dep, extras):
    if dep.package not in source._packages:
        source._versions_for(dep.package, source.convert_dependency(dep).constraint)
    versions = [
        k for k, v in source._packages[dep.package][extras].items() if v is not None
    ]
    return versions[-1]


def _recurse_dependencies(
    source, decision_packages, dependencies, tree_root, tree_parent
):
    packages = OrderedDict()
    for dep in dependencies:
        name = dep.name
        extras = dep.package.req.extras
        resolved_version = decision_packages.get(dep.package) or _find_version(
            source, dep, extras
        )

        tree_node = Node(
            name,
            version=resolved_version,
            parent=tree_parent,
            # pip_string in metadata might be the wrong one (populated differently beforehand, higher up in the tree)
            # left here in case e.g. versions_available is needed in rendered tree
            metadata=source._packages_metadata[name][str(resolved_version)],
            pip_string=dep.pip_string,
        )

        # detect cyclic depenencies
        matches = findall_by_attr(tree_root, name)
        if matches and matches[0] in tree_node.ancestors:
            logger.warning(
                "Cyclic dependency found: %s depends on %s and vice versa.",
                tree_node.name,
                tree_parent.name,
            )
            setattr(tree_node, "cyclic", True)
            packages[(name, str(resolved_version))] = {}
            continue

        packages[(name, str(resolved_version))] = _recurse_dependencies(
            source,
            decision_packages,
            source.dependencies_for(dep.package, resolved_version),
            tree_root,
            tree_node,
        )
    return packages


def build_tree(source, decision_packages):
    tree_root = Node("__root__")
    exhaustive = _recurse_dependencies(
        source, decision_packages, source._root_dependencies, tree_root, tree_root
    )
    return flatten(exhaustive), tree_root


def render_tree(root_tree, max_depth):
    output = []
    for child in root_tree.children:
        lines = []
        for fill, _, node in RenderTree(child):
            if max_depth and node.depth > max_depth:
                continue
            lines.append(
                u"{}{} ({}{})".format(
                    fill,
                    node.pip_string,
                    node.version,
                    ", cyclic" if hasattr(node, "cyclic") else "",
                )
            )
        output += lines
    return "\n".join(output)


@click.command(
    context_settings={"help_option_names": ["-h", "--help"], "max_content_width": 84},
    help="pipgrip is a lightweight pip dependency resolver with deptree preview functionality based on the PubGrub algorithm, which is also used by poetry. For one or more PEP 508 dependency specifications, pipgrip recursively fetches/builds the Python wheels necessary for version solving, and optionally renders the full resulting dependency tree.",
)
@click.argument("dependencies", nargs=-1)
@click.option(
    "--install", is_flag=True, help="Install full dependency tree after resolving.",
)
@click.option(
    "-e", "--editable", is_flag=True, help="Install a project in editable mode.",
)
@click.option(
    "-r",
    "--requirements-file",
    multiple=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="Install from the given requirements file. This option can be used multiple times.",
)
@click.option(
    "--lock", is_flag=True, help="Write out pins to './pipgrip.lock'.",
)
@click.option(
    "--pipe",
    is_flag=True,
    help="Output space-separated pins instead of newline-separated pins.",
)
@click.option(
    "--json",
    is_flag=True,
    help="Output pins as json dict instead of newline-separated pins.",
)
@click.option(
    "--tree", is_flag=True, help="Output human readable dependency tree (top-down).",
)
@click.option(
    "--reversed-tree",
    is_flag=True,
    help="Output human readable dependency tree (bottom-up).",
)
@click.option(
    "--max-depth",
    type=click.INT,
    default=-1,
    help="Maximum tree rendering depth (defaults to -1).",
)
@click.option(
    "--cache-dir",
    envvar="PIP_CACHE_DIR",
    type=click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    default=os.path.join(USER_CACHE_DIR, "wheels", "pipgrip"),
    help="Use a custom cache dir.",
)
@click.option(
    "--no-cache-dir",
    # envvar='PIP_NO_CACHE_DIR',  #  this would be counter-intuitive https://github.com/pypa/pip/issues/2897#issuecomment-231753826
    is_flag=True,
    help="Disable pip cache for the wheels downloaded by pipper. Overrides --cache-dir.",
    # alternatively https://click.palletsprojects.com/en/7.x/options/#boolean-flags
)
@click.option(
    "--index-url",
    envvar="PIP_INDEX_URL",
    default="https://pypi.org/simple",
    help="Base URL of the Python Package Index (default https://pypi.org/simple).",
)
@click.option(
    "--extra-index-url",
    envvar="PIP_EXTRA_INDEX_URL",
    help="Extra URLs of package indexes to use in addition to --index-url.",
)
@click.option(
    "--pre",
    is_flag=True,
    help="Include pre-release and development versions. By default, pip only finds stable versions.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Control verbosity: -v will print cyclic dependencies (WARNING), -vv will show solving decisions (INFO), -vvv for development (DEBUG).",
)
def main(
    dependencies,
    requirements_file,
    install,
    editable,
    lock,
    pipe,
    json,
    tree,
    reversed_tree,
    max_depth,
    cache_dir,
    no_cache_dir,
    index_url,
    extra_index_url,
    pre,
    verbose,
):
    if verbose == 0:
        logger.setLevel(logging.ERROR)
    if verbose == 1:
        logger.setLevel(logging.WARNING)
    if verbose == 2:
        logger.setLevel(logging.INFO)
    if verbose >= 3:
        logger.setLevel(logging.DEBUG)
        logger.debug(str(default_environment()))

    if sum((pipe, json, tree, reversed_tree)) > 1:
        raise click.ClickException("Illegal combination of output formats selected")
    if max_depth == 0 or max_depth < -1:
        raise click.ClickException("Illegal --max_depth selected: {}".format(max_depth))
    if max_depth == -1:
        max_depth = 0
    elif max_depth and not (tree or reversed_tree):
        raise click.ClickException(
            "--max-depth has no effect without --tree or --reversed-tree"
        )
    if requirements_file:
        if dependencies:
            raise click.ClickException(
                "-r can not be used in conjunction with directly passed requirements"
            )
        dependencies = []
        for path in requirements_file:
            dependencies += read_requirements(path)
    if editable:
        if not install:
            raise click.ClickException("--editable has no effect without --install")
        if not sorted(dependencies)[0].startswith("."):
            raise click.ClickException(
                "--editable does not accept input '{}'".format(" ".join(dependencies))
            )
    for dep in dependencies:
        if os.sep in dep:
            raise click.ClickException(
                "'{}' looks like a path, and is not supported yet by pipgrip".format(
                    dep
                )
            )

    if reversed_tree:
        tree = True
    if no_cache_dir:
        cache_dir = tempfile.mkdtemp()

    try:
        source = PackageSource(
            cache_dir=cache_dir,
            index_url=index_url,
            extra_index_url=extra_index_url,
            pre=pre,
        )
        for root_dependency in dependencies:
            source.root_dep(root_dependency)

        solver = VersionSolver(source)
        solution = solver.solve()

        decision_packages = OrderedDict()
        for package, version in solution.decisions.items():
            if package == Package.root():
                continue
            decision_packages[package] = version

        logger.debug(decision_packages)

        packages, root_tree = build_tree(source, decision_packages)

        if lock:
            with io.open(
                os.path.join(os.getcwd(), "pipgrip.lock"), mode="w", encoding="utf-8"
            ) as fp:
                # a lockfile containing `.` will break pip install -r
                fp.write(
                    "\n".join(
                        [
                            "==".join(x)
                            for x in packages.items()
                            if not x[0].startswith(".")
                        ]
                    )
                    + "\n"
                )

        if reversed_tree:
            raise NotImplementedError()
            # TODO root_tree = reverse_tree(root_tree)
        if tree:
            output = render_tree(root_tree, max_depth)
        elif pipe:
            output = " ".join(["==".join(x) for x in packages.items()])
        elif json:
            output = dumps(packages)
        else:
            output = "\n".join(
                [
                    "==".join(x) if not x[0].startswith(".") else x[0]
                    for x in packages.items()
                ]
            )
        click.echo(output)

        if install:
            install_packages(
                # sort to ensure . is added right after --editable
                packages=sorted(dependencies),
                constraints=sorted(
                    "==".join(x) for x in packages.items() if not x[0].startswith(".")
                ),
                index_url=index_url,
                extra_index_url=extra_index_url,
                pre=pre,
                cache_dir=cache_dir,
                editable=editable,
            )
    except (SolverFailure, click.ClickException, CalledProcessError) as exc:
        raise click.ClickException(str(exc))
    # except Exception as exc:
    #     logger.error(exc, exc_info=exc)
    #     raise click.ClickException(str(exc))
